/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifdef WIN32
#pragma warning(disable : 4996)
#include <windows.h>
#define sleep(sec) Sleep(sec*1000)
#include "service.h"
#else
#include <unistd.h>
#endif

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"
#include "install.h"
#include "ciao.h"
#include "cfw.h"
#include "iao_threads.h"

#define RETRY_TIMEOUT 60
#define RETRY_TIMEOUT_URGENT 3

static int shouldconfig=0;
static int configchanged=0;
static time_t nxtkeepalive;
static iao_mutex  keepalivemutex;
static iao_agentinfo agent;
static agentmetadata agentmeta;

char *agentname()
{
	return agentmeta.name;
}
char *AgentName()
{
	return agentmeta.Name;
}

void extract_dir(char *exe, char *result)
{
	char * last_slash;
	strcpy(result, exe);
	last_slash = strrchr(result, '/');
	if (last_slash!=NULL)
	{
		*(last_slash+1) = 0;
	}
	else 
	{
		strcpy(result, "./");
	}
}

void extract_file(char *exe, char *result)
{
	char *last_slash = strrchr(exe, '/');
	if (last_slash!=NULL)
	{
		strcpy(result,last_slash+1);
	}
	else 
	{
		strcpy(result, exe);
	}
#ifdef WIN32
	if (stricmp(result+(strlen(result)-4), ".exe")!=0)
	{
		strcat(result, ".exe");
	}
#endif
}

int save_id(char *dir, char *id)
{
	char fname[MAX_PATH];
        FILE *fd;
        sprintf(fname, "%s%s.id", dir, agentmeta.name );

	fd = fopen(fname, "w");
	if (fd==NULL)
	{
		return -1;
	}
	else
	{
		int ret = fputs(id, fd);;
		fclose(fd);
		return ret==0?0:-1;
	}
}

int obtain_id(char *dir, iao_agentinfo *agent, int skaa)
{
	char buff[IAOP_MAX_LEN];
	int ret;

	ret = iao_register(agent, buff, skaa);
	if (ret==0)
	{
		save_id(dir, buff);
		strcpy(agent->id, buff);
		iao_logmsg("obtain_id: instance id obtained");
		informok();
	}
	return ret;
}
int load_agent_id(char *dir, char *agentid)
{
	char fname[MAX_PATH];
	FILE *fd;
	
	sprintf(fname, "%s%s.id", dir, agentmeta.name);
	fd = fopen(fname, "rt");
	if (fd==NULL)
	{
		return -1;
	}
	else 
	{
		char *fgs = fgets(agentid, IAOP_MAX_LEN, fd);
		fclose(fd);
		if (fgs==NULL)
		{
			return -1;
		}
	}
	return 0;
}

int getandupgrade()
{
	void *buff;
	int len;

	iao_logmsg("getandupgrade: Got a request to upgrade");

	if (iao_getbinary(&agent, &buff, &len)==0)
	{	
		iao_logmsg("getandupgrade: Binary obtained. upgrading...");
		upgrade(buff, len);
	}
	return -1;
}

time_t nextkeepalive()
{
        time_t val;
        iao_lockmutex(keepalivemutex);
        val = nxtkeepalive;
        iao_releasemutex(keepalivemutex);
        return val;
}
void setnextkeepalive(int seconds)
{
        iao_lockmutex(keepalivemutex);
        nxtkeepalive = time(NULL)+seconds;
        iao_releasemutex(keepalivemutex);
}


int act(char *verb, char *x)
{
	if (strcmp(verb, "sleep")==0)
	{
		return atoi(x);
	}
	else if (strcmp(verb, "die")==0)
	{
		iao_logmsg("act: Got an explicit request to die");
		iao_logmsg(x);
		return -1;
	}
	else if (strcmp(verb, "upgrade")==0)
	{
		getandupgrade();
		iao_logmsg("act: Upgrade failed");
		sleep(RETRY_TIMEOUT);
		return 0;
	}
	else if (strcmp(verb, "uninstall")==0)
	{
		uninstall();
		iao_logmsg("act: Uninstall failed");
		return -1;
	}
	else if (strcmp(verb, "config")==0)
	{
		shouldconfig = 1;
		return atoi(x);
	}
	else
	{
		iao_logmsg("act: Unknown verb");
		iao_logmsg(verb);
		return -1;
	}
}

void configure()
{
	if (shouldconfig) 
	{
		if (iao_fetchconfig(&agent, agentmeta.agentparams, agentmeta.resourceparams)!=0)
		{
			iao_logmsg("configure: Failed getting configuration.");
			sleep(3);
		}
		else
		{
			iao_logmsg("configure: Configuration downloaded");
			shouldconfig = 0;
			configchanged = 1;
		}
	}
}

int mainloop()
{
	char verb[IAOP_MAX_LEN];
	char x[IAOP_MAX_LEN];
	
	int agentsample = agentmeta.agentinit(&agent);
	if (agentsample<0)
	{
		iao_logmsg("Agent initialization failed");
		return -1;
	}
	if (agentsample==0)
	{
		agentsample=5*60; /* five minutes default */
	}

	shouldconfig = (agentmeta.agentparams!=NULL||agentmeta.resourceparams!=NULL);

	while (1)
	{
		int ret;
		int rest = agentsample;

		configure();
		
		ret = agentmeta.agentmain(&agent, verb, x, 
			configchanged?MODE_CONFIG_CHANGED:0);
		if (ret==0)
		{
			configchanged = 0;
			ret = act(verb, x) ;
			if (ret<0)
			{
				return -1;
			}
			setnextkeepalive(ret);
		}
		if (ret==1)
		{
			configchanged = 0;
		}
		if (ret==-1)
		{
			rest = 60;
		}
		sleep(shouldconfig?1:rest);
	}
}

void openmessage()
{
	char buff[256];
	sprintf(buff, "------ %s launched -------", agentmeta.Name );
	iao_logmsg(buff);
}

void setexename(char *exename)
{
	bs2s(exename);
	extract_dir(exename, getdir() );
	extract_file(exename, getexefile() );
	openmessage();
}


void keepalivethread(void *param)
{
	char verb[IAOP_MAX_LEN];
	char x[IAOP_MAX_LEN];

	int urgent_retries = 0;
	int rest;
	while (1)
	{
		if (time(NULL)>=nextkeepalive())
		{
			int ret = iao_keepalive(&agent, verb, x);
			if (ret==0)
			{
				urgent_retries = 10;
				ret = act(verb, x);
				if (ret<0)
				{
					exit(-1);
				}
				setnextkeepalive(ret);
				rest = ret;
			}
			else
			{
                        	rest =  urgent_retries-- >0 ? RETRY_TIMEOUT_URGENT : RETRY_TIMEOUT;
			}
		}
		else
		{
			rest = (int) (nextkeepalive() - time(NULL));
		}
		sleep(rest);
	}
}

int init(char *supplied_account_id)
{
	if (iao_init()!=0)
	{
		iao_logmsg("Can't initialize ciao library");
		return -1;
	}

	/* Variables */
	strcpy(agent.type, agentmeta.type);
	strcpy(agent.version, agentmeta.version);

	/* fetch account id - when writing a private agent you should get
	   this from the user (copy from the account details) */
	if (supplied_account_id==NULL)
	{
		if (iao_getembeddedvariable("ACCOUNT_ID", agent.id)==-1)
		{
			iao_logmsg("init: Can't find ACCOUNT_ID in embedded variables."
                                   " You should get a new agent version.");
			return -1;
		}
	}
	else
	{
		strcpy(agent.id, supplied_account_id);
	}

	/* Obtain agent id */
	if (load_agent_id(getdir(), agent.id)!=0)
	{
		while (obtain_id(getdir(), &agent, 0)!=0)
        	{
			sleep(RETRY_TIMEOUT);
		}
		if (agentmeta.agentregister!=NULL)
		{
			agentmeta.agentregister(&agent);
		}
	}
	keepalivemutex = iao_createmutex();
 	setnextkeepalive(10);/*give 10 seconds for agent imp */

	iao_spawnthread(keepalivethread, NULL);
	
	return mainloop();
}

#ifdef WIN32

int main(int argc, char *argv[])
{
	agentmeta = getagentinfo();
	setexename(argv[0]);
	if (argc>1)
	{
#ifdef CIAO_TEST
		if (strcmp(argv[1], "account")==0)
		{
			init(argv[2]);
			return 0;
		}
#endif
		if (strcmp(argv[1],"service")==0)
		{
			iao_logmsg("Starting agent service...");
			ServiceMain();
			return 0;
		}
		else if (strcmp(argv[1], "uninstall")==0)
		{
			UninstallService();
			iao_logmsg("Uninstalled");
			return 0;
		} 
	}

	if (install()==-1)
	{
		if (ask("Installation has failed. Would you like to see the log file?"))
		{
			showlog();
		}
	}
	return 0;

}
#else
int main(int argc, char* argv[])
{
	agentmeta = getagentinfo();
	setexename(argv[0]);

	if (argc>=2)
	{
		if (strcmp(argv[1], "install")==0)
		{
			return install();
		}
		if (strcmp(argv[1], "stop")==0)
		{
			return 0;
		}
		if (strcmp(argv[1], "account")==0)
		{
			if (argc>=3)
			{
				return init(argv[2]);
			}
		}
		printf("Unknown argument: %s\n", argv[1]);
		return -1;
	}
	return init(NULL);
}
#endif
