/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <windows.h>
#pragma warning(disable:4996)
#endif

#include "ciao.h"

typedef struct iao_configpair_struct {
	char name[256];
	char value[1024];
} iao_configpair;
typedef struct iao_configpairs_struct {
	iao_configpair *items;
	int size;
} iao_configpairs;
typedef struct iao_configresource_struct {
	char resourcename[512];
	iao_configpairs pairs;
} iao_configresource;

typedef struct iao_configresources_struct {
	iao_configresource *items;
	int size;
} iao_configresources;
typedef struct iao_config_struct {
	iao_configpairs pairs;
	iao_configresources resources;
} iao_config;

static iao_config agentconfig;
static char *resourcelist = NULL;

static char escapechar(char *s, int len)
{
    typedef struct { char *seq; char chr; int len;} seqt;
    seqt seq[] = {
        {"amp",'&', 3},
        {"quot", '"', 4},
        {"lt", '<', 2},
        {"gt", '>', 2},
        {NULL, 0, 0}};
        seqt *p = seq;

        while (p->seq!=NULL)
        {
            if (p->len==len && strncmp(p->seq, s, len)==0)
            {
                return p->chr;
            }
            p++;
        }
        return 0;
}

static void xmlunescape(char *buff)
{
    char *last = NULL;
    char *p = buff;
    char *c = buff;

    while (*p)
    {
        char chr;
        if (*p==';' && last!=NULL && p-last<6 && 
            (chr = escapechar(last+1, (int)(p-last-1)))!=0)
        {       
            c = c - (p-last);
            *c = chr;
            last = NULL; 
        }       
        else    
        {       
            if (*p=='&')
            {       
                last = p;
            }
            *c = *p;
        }
        c++;
        p++;
    }
    *c = 0;
}


/*static void debugconfig()
{
        int i;
        printf("---START CONFIG---\n");
        for (i = 0;i<agentconfig.pairs.size;++i)
        {
                printf("%s=%s\n", agentconfig.pairs.items[i].name, agentconfig.pairs.items[i].value);
        }
        for (i = 0; i<agentconfig.resources.size;++i)
        {
                int j;
                printf("%s\n", agentconfig.resources.items[i].resourcename);
                for (j = 0;j<agentconfig.resources.items[i].pairs.size;++j)
                {
                        printf("\t%s=%s\n",
                                agentconfig.resources.items[i].pairs.items[j].name,
                                agentconfig.resources.items[i].pairs.items[j].value);
                }
        }
        printf("---END CONFIG---\n");
}*/

static void freepairs(iao_configpairs pairs)
{
	if (pairs.items!=NULL) {
		free(pairs.items);
		pairs.items=NULL;
		pairs.size = 0;
	}
}

static void freeconfig()
{
	int i;

	freepairs(agentconfig.pairs);
	agentconfig.pairs.size = 0;

	if (agentconfig.resources.items!=NULL)
	{
		for (i=0;i<agentconfig.resources.size;++i)
		{
			freepairs(agentconfig.resources.items[i].pairs);
		}
		free(agentconfig.resources.items);
		agentconfig.resources.size = 0;
		agentconfig.resources.items = NULL;
	}
	if (resourcelist!=NULL)
	{
		free(resourcelist);
		resourcelist = NULL;
	}
}

static char *pairsgetvalue(iao_configpairs pairs, char *name)
{
	int i;
	for (i=0;i<pairs.size;++i)
	{
		if (strcmp(name, pairs.items[i].name)==0)
		{
			return pairs.items[i].value;
		}
	}
	return NULL;
}

static int peelnode(char *name, char *buff, int len)
{
	char o[256];
	char c[256];
	int clen;
	int olen;
	
	sprintf(o, "<%s>", name);
	sprintf(c, "</%s>", name);
	olen = (int) strlen(o);
	clen = (int) strlen(c);

	if (memcmp(buff, o, olen)!=0)
	{
		sprintf(c, "peelnode: Can't find opening tag: %s", o);
		iao_logmsg(c);
		return -1;
	}
	if (memcmp(buff+len-clen, c, clen)!=0)
	{
		sprintf(o, "peelnode: Can't find closing tag: %s", c);
		iao_logmsg(o);
		return -1;
	}
	len = len - olen - clen;
	memmove(buff, buff+olen, len);
	*(buff+len)=0;
	return len;
}


static int parsepair(iao_configpair *pair, char *buff)
{
	int len;
	char closingtag[256];
	char *closetag;
	char *openend;

	if (*buff!='<')
	{
		iao_logmsg("parsepair: buff not a node beginning");
		return -1;
	}
	openend = strstr(buff, ">");

	if (openend==NULL)
	{
		iao_logmsg("parsepair: Parse error couldn't find > of opening tag");
		return -1;
	}

	len = (int)(openend - buff -1);
	strncpy(pair->name, buff+1, len);
	pair->name[len]=0;

	sprintf(closingtag, "</%s>", pair->name);
	closetag = strstr(openend+1, closingtag);
	if (closetag==NULL)
	{
		char msg[256];
		sprintf(msg, "parsepair: Can't find closing tag \"%s\"", closingtag);
		iao_logmsg(msg);
		return -1;
	}
	len = (int) (closetag-openend-1);
	strncpy(pair->value, openend+1, len);
	pair->value[len]=0;
	xmlunescape(pair->value);
	return (int)(closetag-buff+strlen(closingtag));
}

static int parsepairs(iao_configpairs *pairs, char *buff, int len)
{
	iao_configpair parsedpairs[256];
	int foundpairs=0;
	int pos=0;

	while (pos<len)
	{
		int ret = parsepair(parsedpairs+(foundpairs++), buff+pos);
		if (ret==-1)
		{
			return -1;
		}
		else
		{
			pos+=ret;
		}
	}

	if (foundpairs==0)
	{
		pairs->items = NULL;
	}
	else
	{
		len = sizeof(iao_configpair)*foundpairs;
		pairs->items = malloc(len);
		memcpy(pairs->items, parsedpairs, len);
	}
	pairs->size = foundpairs;
	return pos;
}

static int parseresource(char *p, iao_configresource *resource)
{
	char *opentag = "<resource>";
	char *closetag = "</resource>";
	char *begin;
	char *end;
	int len;

	begin = strstr(p, opentag);
	end = strstr(p, closetag);
	if (begin==NULL || end==NULL)
	{
		iao_logmsg("parseresource: Couldn't find <resource></resource> pair");
		return -1;
	}

	len = parsepairs(&resource->pairs, begin+strlen(opentag), (int)(end-begin-strlen(opentag)) );
	if (len==-1)
	{
		iao_logmsg("parseresource: Failed parsing resource");
		return -1;
	}

	if (resource->pairs.size>0)
	{
		strcpy(resource->resourcename, resource->pairs.items[0].value);
	}
		
	return (int)(end-begin+strlen(closetag));
}

static int parseresources(char *p)
{
	iao_configresource resources[256];
	int parsedresources=0;

	if (peelnode("resources", p, (int)strlen(p))==-1)
	{
		return -1;
	}

	while (*p)
	{
		int len = parseresource(p, &resources[parsedresources++]);
		if (len==-1)
		{
			return -1;
		}
		p+=len;
	}

	if (parsedresources==0)
	{
		agentconfig.resources.items=NULL;
	}
	else
	{
		int len = parsedresources*sizeof(iao_configresource);
		agentconfig.resources.items=malloc(len);
		memcpy(agentconfig.resources.items, resources, len);
	}
	agentconfig.resources.size = parsedresources;
	return 0;
}

static void buildresourcelist()
{
	int len = 0;
	char *p;
	int i;
	for (i=0;i<agentconfig.resources.size;++i)
	{
		len+=(int)(strlen(agentconfig.resources.items[i].resourcename)+1);
	}
	if (len==0)
	{
		return;
	}

	resourcelist = malloc(len+1);
	p = resourcelist;
	for (i=0;i<agentconfig.resources.size;++i)
	{
		int len = sprintf(p, "%s",
                             agentconfig.resources.items[i].resourcename);
		p+=(len+1);
	}
	*p = 0;
}

int iao_parseconfig(char *buff, int len)
{
	char *p;

	freeconfig();

	len = peelnode("config", buff, len);
	if (len==-1)
	{
		return -1;
	}

	/* Find end of pairs */
	p = strstr(buff, "<resources>");
	
	len = p==NULL?len:(int)(p-buff);
	
	len = parsepairs(&agentconfig.pairs, buff, len);
	if (len==-1) 
	{
		return -1;
	}
	if (p!=NULL)
	{
		if (parseresources(p)==-1)
		{
			return -1;
		}
		buildresourcelist();
	}
	/*debugconfig();*/
	return 0;
}

char *iao_configvalue(char *name)
{
	return pairsgetvalue(agentconfig.pairs, name);
}


char *iao_configresourcevalue(char *resource, char *name)
{
	int i;
	for (i=0;i<agentconfig.resources.size;++i)
	{
		if (strcmp(resource, agentconfig.resources.items[i].resourcename)==0)
		{
			return pairsgetvalue(agentconfig.resources.items[i].pairs, name);
		}
	}
	return NULL;
}

char *iao_configresourcelist()
{
	return resourcelist;
}

void iao_configinit()
{
	agentconfig.resources.items = NULL;
	agentconfig.resources.size = 0;
	agentconfig.pairs.items = NULL;
	agentconfig.pairs.size = 0;
	resourcelist = NULL;
}
