/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#include "ciao.h"
#include "cfw.h"
#include "onping.h"
#include <string.h>
#include <stdio.h>
#include <time.h>

#ifdef WIN32
#include <windows.h>
#pragma warning(disable:4996)
#endif

int agentinit(iao_agentinfo *agent);
int agentmain(iao_agentinfo *agent, char *verb, char *x, int mode);

agentmetadata getagentinfo()
{
	static agentmetadata meta;
	static iao_confparam resourceparams[] = {
		{ "host",  "String", "" },
		{ "retries",  "Integer", "0" },
		{0}
	};

	static iao_confparam agentparams[] = {
		{ "Timeout",  "Integer(Milliseconds)", "250" },
		{0}
	};

	meta.Name = "OnPing";
	meta.name = "onping";
	/* When writing your own private agent don't add the ruon prefix to its type name. */
	/* This way you are certain to maintain name uniqueness. */
	meta.type = "ruon.OnPing";
        meta.version = "1.6";
	
	meta.agentparams = agentparams;;
	meta.resourceparams = resourceparams;
	meta.agentregister = NULL;
	meta.agentinit = agentinit;
	meta.agentmain = agentmain;

	return meta;
}



int agentinit(iao_agentinfo *agent)
{
	if (pinginit()!=0)
	{
		return -1;
	}
	return 60;
}


static hoststate hosts[MAX_HOSTS];
static int size = 0;
static int timeout = 250;

static void checkconfig()
{
	/* Reset configuration */
	char *p = iao_configresourcelist();
	timeout = atoi(iao_configvalue("Timeout"));
	size = 0;
	if (p!=NULL)
	{
		while (*p!=0 && size<MAX_HOSTS)
		{
			int len = sprintf(hosts[size].host, "%s", p);
			hosts[size].retries = atoi(iao_configresourcevalue(p, "retries"));
			hosts[size].failed = 0;
			hosts[size].reported = 0;
			size++;
			p+=(len+1);
		}
	}
}


char *retcode2id(int code)
{
	switch (code)
	{
	case HOST_RESOLVE_ERROR:
		return "UNKNOWNHOST";
	case ECHO_CORRUPT:
		return "ICMP_CORRUPT";
	case QTIMEOUT:
		return "ICMP_QTIMEOUT";
	case TTL_EXPIRED:
		return "ICMP_TTL";
	case PING_FAILED:
	default:
		return "ICMP_NOECHO";
	}
}

static int report(iao_agentinfo *agent, char *verb, char *x)
{
	iao_alarm alarms[MAX_HOSTS];
	int ac = 0;
	int i;
	int ret;

	for (i=0;i<size;++i)
	{
		if (hosts[i].actual==0) {
			hosts[i].failed = 0;
		} else {
			if (++hosts[i].failed <= hosts[i].retries) {
				char buff[256];
				sprintf(buff, "%s failed %d times with %d retries: no alarm", hosts[i].host, hosts[i].failed, hosts[i].retries);
				iao_logmsg(buff);
				hosts[i].actual = 0;
			}
		}
		if ( (hosts[i].actual==0 && hosts[i].reported!=0) || (hosts[i].actual!=0 && hosts[i].failed>hosts[i].retries))
		{
			strcpy(alarms[ac].resource, hosts[i].host);
			strcpy(alarms[ac].description, hosts[i].msg);
			alarms[ac].type = 'A';

			switch (hosts[i].actual)
			{
				case OKAY: /* ok */
					alarms[ac].type = 'C';
					strcpy(alarms[ac].id, retcode2id(hosts[i].reported));
					break;
				case HOST_RESOLVE_ERROR: /* unknown host */
					strcpy(alarms[ac].id, retcode2id(hosts[i].actual));
					alarms[ac].severity = 'M';
					break;
				case ECHO_CORRUPT: /* packet corruption */
					strcpy(alarms[ac].id, retcode2id(hosts[i].actual));
					alarms[ac].severity = 'M';
					break;
				case QTIMEOUT:
					strcpy(alarms[ac].id, retcode2id(hosts[i].actual));
					alarms[ac].severity = 'M';
					break;
				case TTL_EXPIRED:
                                        strcpy(alarms[ac].id, retcode2id(hosts[i].actual));
					strcpy(alarms[ac].description, "The time to live (TTL) expired in transit");
                                        alarms[ac].severity = 'M';
                                        break;
				case PING_FAILED: /* no echo */
				default:
					strcpy(alarms[ac].id, retcode2id(hosts[i].actual));
					alarms[ac].severity = 'C';
					break;
			}
			ac++;
		}
	}
	ret = iao_alarms(agent, alarms, ac, 0, verb, x);
	if (ret==0)
	{
		for (i=0;i<size;i++)
		{
			hosts[i].reported = hosts[i].actual;
		}
	}
	return ret;
}

int agentmain(iao_agentinfo *agent, char *verb, char *x, int mode)
{
	int configchanged = mode&MODE_CONFIG_CHANGED;

	if (configchanged)
	{
		checkconfig();
	}
	if (checkping(hosts,size,timeout, 3000) || configchanged)
	{
		return report(agent, verb, x);
	}
	else
	{
		return 1;
	}
}
