/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#ifdef WIN32

#include <windows.h>
#include <Ipexport.h>
#include <Icmpapi.h>
#include <stdio.h>

#include "ciao.h"
#include "onping.h"

#pragma warning(disable:4996)

#define dll1 "iphlpapi.dll"
#define dll2 "icmp.dll"

static HANDLE hIcmpFile;
static int timeout1 = 3000;
static int timeout2 = 250;

static HANDLE (WINAPI *IcmpCreateFileP)(void) = NULL;
static DWORD (WINAPI *IcmpSendEcho2P)(HANDLE, HANDLE, FARPROC, PVOID, IPAddr, LPVOID,
			 WORD, PIP_OPTION_INFORMATION, LPVOID, DWORD, DWORD) = NULL;

static FARPROC loaddllmethod(HMODULE *module, char *dllname, char *method)
{
		*module = LoadLibrary(dllname);
		if (*module!=NULL)
		{
			return GetProcAddress(*module, method);
		}
		else
		{
			return NULL;
		}
}

static void loadmethod(char *method, void **function)
{
	static HMODULE dll = NULL;
	if (dll==NULL)
	{
		*function = loaddllmethod(&dll, dll1, method);
		if (*function==NULL)
		{
			*function = loaddllmethod(&dll, dll2, method);
		}
	}
	else
	{
		*function = GetProcAddress(dll, method);
	}
}

static void initmethods()
{
	loadmethod("IcmpSendEcho2", (void**)&IcmpSendEcho2P);
	loadmethod("IcmpCreateFile", (void**)&IcmpCreateFileP);
}

static IPAddr netaddr(char *s)
{
	IPAddr ret = inet_addr(s);
	if (ret==INADDR_NONE)
	{
		struct hostent *hp = gethostbyname(s);
		if (hp!=NULL)
		{
			ret = *((long*)hp->h_addr); 
		}
	}
	return ret;
}



void SystemError(char *buff)
{
	LPVOID lpMsgBuf;
	size_t len;
	FormatMessageA( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),/* Default language */
		(LPSTR) &lpMsgBuf,
		0,
		NULL 
		);
	len = strlen(lpMsgBuf);
	len = len>1000?999:len-2;
	strncpy(buff, (char*)lpMsgBuf, len);
	buff[len]=0;
	LocalFree( lpMsgBuf );
}


void sysmsg(char *retBuff, char *errmsg)

{
	char buff[1024];

	SystemError(buff);
	sprintf(retBuff, "%s: %s", errmsg, buff);
}

static char *ipheaderErrString(ULONG Status)
{
    switch (Status)
    {
    case IP_SUCCESS:
        return "The status was success";
    case IP_BUF_TOO_SMALL:
        return "The reply buffer was too small";
    case IP_DEST_NET_UNREACHABLE:
        return "The destination network was unreachable";
    case IP_DEST_HOST_UNREACHABLE:
        return "The destination host was unreachable";
    case IP_DEST_PROT_UNREACHABLE:
        return "The destination protocol was unreachable"; 
    case IP_DEST_PORT_UNREACHABLE:
        return "The destination port was unreachable"; 
    case IP_NO_RESOURCES:
        return "Insufficient IP resources were available";
    case IP_BAD_OPTION:
        return "A bad IP option was specified";
    case IP_HW_ERROR:
        return "A hardware error occurred";
    case IP_PACKET_TOO_BIG:
        return "The packet was too big";
    case IP_REQ_TIMED_OUT:
        return "The request timed out";
    case IP_BAD_REQ:
        return "A bad request";
    case IP_BAD_ROUTE:
        return "A bad route";
    case IP_TTL_EXPIRED_TRANSIT:
        return "The time to live (TTL) expired in transit"; 
    case IP_TTL_EXPIRED_REASSEM:
        return "The time to live expired during fragment reassembly"; 
    case IP_PARAM_PROBLEM:
        return "A parameter problem"; 
    case IP_SOURCE_QUENCH:
        return "Datagrams are arriving too fast to be processed and datagrams may have been discarded"; 
    case IP_OPTION_TOO_BIG:
        return "An IP option was too big"; 
    case IP_BAD_DESTINATION:
        return "A bad destination"; 
    case IP_GENERAL_FAILURE:
        return "A general failure. This error can be returned for some malformed ICMP packets"; 
    default:
        return "Unknown Status Result";
    }
}

int ping(char *host, int milisec, char *errmsg, int *rtt, int ttl)
{
	IPAddr ip;
	char *data = "A_GOOD_LUCK_PACKET"; /* For a lucky network */
	int datalen = (int)strlen(data);
	char replybuff[1024];
    IP_OPTION_INFORMATION ipOptionInfo;

	ip = netaddr(host);
	if (ip==INADDR_NONE)
	{
		strcpy(errmsg, "Can't translate host/ip");
		return HOST_RESOLVE_ERROR;
	}

    if (ttl>=0)
    {
        ipOptionInfo.Ttl = ttl;
        ipOptionInfo.Tos = 0;
        ipOptionInfo.Flags = 0;
        ipOptionInfo.OptionsSize = 0;
        ipOptionInfo.OptionsData = NULL;
    }
    
    
    if (IcmpSendEcho2P(hIcmpFile, NULL, NULL, NULL, ip, data, (WORD)datalen, 
		ttl==-1?NULL:&ipOptionInfo, replybuff, 1024, milisec))
	{
        ICMP_ECHO_REPLY *header = (ICMP_ECHO_REPLY *)replybuff;
        if (header->Status==IP_SUCCESS)
        {
		    char *echo = (char*)replybuff+sizeof(ICMP_ECHO_REPLY);
		    if (strncmp(echo, data, datalen)==0)
		    {
			    strcpy(errmsg,"Ping reestablished within timeout");
			    (*rtt) = ((ICMP_ECHO_REPLY*)replybuff)->RoundTripTime;
			    return 0;
		    }
		    else
		    {
                strcpy(errmsg, "Return packet received but it was corrupted");
			    return ECHO_CORRUPT;
            } 
        }
        else if (ttl>=0 && header->Status==IP_TTL_EXPIRED_TRANSIT) 
        {
            sprintf(errmsg, "%d.%d.%d.%d", header->Address&0xff, (header->Address&0xff00)>>8, (header->Address&0xff0000)>>16, (header->Address&0xff000000)>>24);
            return TTL_EXPIRED;
        }
        else
        {
            strcpy(errmsg, ipheaderErrString(header->Status));
            return PING_FAILED;
        }
	}
	else 
	{
		if (GetLastError()==11010) 
		{
			sprintf(errmsg, "Ping timed out after %d seconds", milisec/1000);
		}
		else
		{
			sysmsg(errmsg, "Ping failed");
		}
		return PING_FAILED;
	}
	return 0;
}

int pinginit()
{
	initmethods();
	hIcmpFile= IcmpCreateFileP();
	if (hIcmpFile==INVALID_HANDLE_VALUE)
	{
		iao_logsysmsg("Can't open ICMP context");
		return -1;
	}
	return 0;
}

int pingretry(char *host, int timeout1, int timeout2, char *errmsg)
{
	int retries=3;
	int ret;
	

	while (retries>0)
	{
		int rtt;
		ret = ping(host, timeout2, errmsg, &rtt, -1);
		switch(ret)
		{
		case OKAY:
			if (rtt>timeout1)
			{
				ret = QTIMEOUT;
				sprintf(errmsg, "Response time of %dms exceeded %dms timeout", rtt, timeout1);
				retries--;
			}
			else
			{
				retries = 0;
			}
			break;
		case ECHO_CORRUPT:
		case PING_FAILED:
			retries--;
			break;
		case HOST_RESOLVE_ERROR:
		default:
			retries=0;
			break;
		}
	}
	return ret;
}

DWORD WINAPI pingthread(hoststate *host)
{
	host->actual = pingretry(host->host, timeout1, timeout2, host->msg);
	return 0;
}



int checkping(hoststate *hosts, int size,int first_timeout, int final_timeout)
{
	HANDLE threads[MAX_HOSTS];
	int i;
	int notif=0;

	timeout1 = first_timeout;
	timeout2 = final_timeout;

	for (i=0;i<size;i++)
	{
		DWORD tid;
		threads[i] = CreateThread(NULL, 64000, (LPTHREAD_START_ROUTINE)pingthread, hosts+i, 0, &tid);
		if (threads[i]==NULL)
		{
			iao_logsysmsg("Failed creating thread");
		}
	}

	for (i=0;i<size;i++)
	{
		WaitForSingleObject(threads[i], INFINITE);
		CloseHandle(threads[i]);

		if (hosts[i].actual!=hosts[i].reported)
		{
			notif = 1;
		}
	}
	return notif;
}
#endif /* WIN32 */
