/*
** Copyright R-U-ON, 2006
** www.r-u-on.com
*/

#include "system.h"

#define MIN_CPU_SAMPLE 20

#ifdef WIN32

#include <windows.h>
#include <pdh.h>
#include <pdhmsg.h>
#include <time.h>
#include <stdio.h>
#include "common.h"
#pragma warning(disable:4996)


int loadvolumes(volume *volumes, int max)
{

	char buff[1024];
	int ret = GetLogicalDriveStringsA(1024, buff);
	if (ret>0)
	{
		int vols = 0;
		char *p = buff;
		while (*p!=0)
		{
			if (GetDriveTypeA(p)==DRIVE_FIXED)
			{
				strcpy(volumes[vols].name, p);
				volumes[vols].alarmed = -1;
				vols++;
			}
			p+=strlen(p)+1;
			if (vols==max) 
			{
				p = 0;
			}

		}
		return vols;
	}
	else
	{
		return -1;
	}
}

int volumeusage(char *p)
{
	ULARGE_INTEGER avialableb, totalb, freeb;

	BOOL ret = GetDiskFreeSpaceExA(p, &avialableb, &totalb, &freeb);
	if (ret)
	{
		unsigned __int64 percent = 0;
		unsigned __int64 hundred = 100;
		if (totalb.QuadPart!=0)
		{
			percent = (totalb.QuadPart -avialableb.QuadPart)*hundred / totalb.QuadPart;
		}
		
		return (int) percent;
		
	}
	else 
	{
		return -1;
	}
}

static int percent(unsigned __int64 total, unsigned __int64 free)
{
		unsigned __int64 occ = total - free;
		unsigned __int64 hundred = 100;
	
		return (int) ((occ*hundred)/total);
}

static void memory(int readings[])
{
	MEMORYSTATUSEX statex;

	statex.dwLength = sizeof (statex);
	if (GlobalMemoryStatusEx (&statex)==0)
	{
		return;
	}

	readings[SWAP] = percent(statex.ullTotalPageFile, statex.ullAvailPageFile);
	readings[PMEM] = statex.dwMemoryLoad;
}


static PDH_HQUERY hQuery=NULL;
static PDH_HCOUNTER cpucounter;
static time_t lastquerytime;


static void cnamebyindex(int idx, char *buff)
{
	DWORD size = 1024;
	if (PdhLookupPerfNameByIndex(NULL, idx, buff, &size)!=ERROR_SUCCESS)
	{
		buff[0] = 0;
	}
}

static char *cpucountername()
{
	static char *name = NULL;
	if (name==NULL)
	{
		int obj = 238;
		int counter = 6;
		char objname[1024];
		char countername[1024];
		DWORD len = PDH_MAX_COUNTER_PATH;
		static char buff[PDH_MAX_COUNTER_PATH];
		PDH_COUNTER_PATH_ELEMENTS cpe;
	
		name = buff;

		cnamebyindex(obj, objname);
		cnamebyindex(counter, countername);

		cpe.szMachineName = NULL;
		cpe.szObjectName = objname;
		cpe.szInstanceName = "_Total";
		cpe.szParentInstance = NULL;
		cpe.dwInstanceIndex = -1;
		cpe.szCounterName = countername;

		PdhMakeCounterPath(&cpe, name, &len, 0);
	}
	return name;
}


static void performanceinit()
{
	if (PdhOpenQuery(NULL, 0, &hQuery) != ERROR_SUCCESS)
	{
		iao_logsysmsg("Error while openning performance query");
		return;
	}
	if (PdhAddCounter(hQuery, cpucountername(), 0, &cpucounter) != ERROR_SUCCESS) 
	{
		iao_logsysmsg("Error adding cpu counter");
		return;
	}
	if (PdhCollectQueryData(hQuery)!=ERROR_SUCCESS)
	{
		iao_logsysmsg("Error collecting performance data");
		return;
	}
}

static void performance(int readings[])
{
	PDH_FMT_COUNTERVALUE countervalue;
	time_t now = time(NULL);
	
	readings[CPU] = 0;
	
	if (hQuery==NULL)
	{
		performanceinit();
		lastquerytime = now;
		return; // The first time, nothing we can do
	}

	if (now-lastquerytime<=MIN_CPU_SAMPLE)
	{
		return; 
	}
	lastquerytime = now;


	if (PdhCollectQueryData(hQuery)!=ERROR_SUCCESS)
	{
		iao_logsysmsg("Error collecting performance data");
		return;
	}

	if (PdhGetFormattedCounterValue(cpucounter, PDH_FMT_LONG, NULL, &countervalue)!= ERROR_SUCCESS) 
	{
		iao_logsysmsg("Error getting formatted value");
		return;
	}


	if (countervalue.CStatus!=ERROR_SUCCESS)
	{
		char buff[256];
		sprintf(buff, "Bad counter CStatus: %d", countervalue.CStatus);
		iao_logsysmsg(buff);
		return;
	}

	readings[CPU] = countervalue.longValue;
	
}



int getmetrics(int readings[])
{
	memory(readings);
	performance(readings);
	return 0;
}

#else
static int percent(int total, int rootavail, int avail)
{
	double t = total-rootavail+avail;
	double a = avail;
	return 100 - (int)(a*100.00/t);
}

#endif /* WIN32 */


#ifdef FreeBSD
#define Darwin
#endif
#ifdef Darwin

#include <string.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <sys/time.h>
#include "common.h"

#ifndef FreeBSD
#include <mach/mach_port.h>
#include <mach/mach_types.h>
#include <mach/mach_host.h>
#include <mach/mach_init.h>
static mach_port_t machport;
#endif

#define CPU_STATES 3

static int interestingfs(struct statfs *s)
{
        char *dev = "/dev/";

        if (strncmp(s->f_mntfromname, dev, strlen(dev))!=0)
        {
                return 0;
        }
	if (s->f_flags&MNT_RDONLY)
	{
		return 0;
	}
        return 1;
}

int loadvolumes(volume *volumes, int max) 
{
	struct statfs buff[max];
	int i;
	int vols = 0;
	
	max = getfsstat(buff, max*sizeof(struct statfs), MNT_NOWAIT);
	if (max==-1)
	{
		return -1;
	}
	
	for (i=0; i<max;++i)
	{
		struct statfs *s = &(buff[i]);
		if (interestingfs(s))
		{
			strncpy(volumes[vols].name, s->f_mntonname, MAX_VOLUME_NAME);
			volumes[vols].name[MAX_VOLUME_NAME-1] = 0;
			volumes[vols].alarmed = -1;
			vols++;
		}
	}
	return vols;
}

int volumeusage(char *p)
{
	struct statfs s;
	int pcnt = 0;
	if (statfs(p, &s)==0)
	{
		pcnt = percent(s.f_blocks, s.f_bfree, s.f_bavail) ;
		return pcnt - 1;
	}
	else
	{
		return -1;
	}
}

static long percentages(int cnt, int *out, long *new, long *old, long *diffs)
{
    register int i;
    register long change;
    register long total_change;
    register long *dp;
    long half_total;

    /* initialization */
    total_change = 0;
    dp = diffs;

    /* calculate changes for each state and the overall change */
    for (i = 0; i < cnt; i++)
    {
        if ((change = *new - *old) < 0)
        {
            /* this only happens when the counter wraps */
            change = (int)
                ((unsigned long)*new-(unsigned long)*old);
        }
        total_change += (*dp++ = change);
        *old++ = *new++;
    }

    /* avoid divide by zero potential */
    if (total_change == 0)
    {
        total_change = 1;
    }

    /* calculate percentages based on overall change, rounding up */
    half_total = total_change / 2l;
    for (i = 0; i < cnt; i++)
    {
        *out++ = (int)((*diffs++ * 1000 + half_total) / total_change);
    }

    /* return the total in case the caller wants to use it */
    return(total_change);
}
#ifndef FreeBSD
static int cpusnapshot()
{
        mach_msg_type_number_t  count;
        host_cpu_load_info_data_t load;

        count = HOST_CPU_LOAD_INFO_COUNT;
        if (host_statistics(machport, HOST_CPU_LOAD_INFO,
            (host_info_t)&load, &count)==KERN_SUCCESS)
        {
                int i;
        	static long cp_time[CPU_STATES];
        	static long cp_old[CPU_STATES];
        	static long cp_diff[CPU_STATES];
        	static int cpu_states[CPU_STATES];

                for (i=0;i<CPU_STATES;++i)
                {
                        cp_time[i] = load.cpu_ticks[i];
                }
                percentages(CPU_STATES, cpu_states, cp_time, cp_old, cp_diff);
		return (1000-cpu_states[2])/10;
        }
        else
        {
                iao_logsysmsg("Error getting CPU load");
                return -1;
        }
}


static int cpuusage()
{
        static time_t last = 0;
        time_t now = time(NULL);

	if (last==0)
	{
		if (cpusnapshot()>=0) /* init */
		{
			last = now;
		}
		return 0;
	}
	else
	{
		if (now-last<MIN_CPU_SAMPLE)
		{
			return 0;
		}
		else
		{
			int ret = cpusnapshot();
			if (ret>=0)
			{
				last = now;
        			return ret;
			}
			else
			{
				return 0;
			}
		}
	}
}


static int physicalmemory()
{
        vm_statistics_data_t vmstat;
        mach_msg_type_number_t  count;

        count = sizeof(vmstat) / sizeof(natural_t);
        if (host_statistics(machport, HOST_VM_INFO,
            (host_info_t)&vmstat, &count)== KERN_SUCCESS)
        {
                unsigned long long totalmem =
                        vmstat.free_count + vmstat.active_count+
                        vmstat.inactive_count + vmstat.wire_count;
                unsigned long long usedmem = 
			totalmem - vmstat.free_count - vmstat.inactive_count;
                return usedmem*100L/totalmem;
        }
        else
        {
                iao_logsysmsg("Error getting physical memory usage");
                return 0;
        }
}
#endif

int getmetrics(int readings[])
{
#ifndef FreeBSD
        static int portinit = 0;
        if (portinit==0)
        {
                machport = mach_host_self();
                portinit = 1;
        }
	readings[CPU] = cpuusage();
	readings[PMEM] = physicalmemory();
#endif
	return 0;
}


#endif /* Darwin */







#ifdef Linux
#include <stdio.h>
#include <string.h>
#include <mntent.h>
#include <sys/statvfs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>

#define NCPUSTATES 4



static int interestingfs(struct mntent *mnt)
{
	char *dev = "/dev/";
	char *readonly = "ro,";

	if (strncmp(mnt->mnt_fsname, dev, strlen(dev))!=0)
	{
		return 0;
	}
	if (strstr(mnt->mnt_opts, readonly)!=NULL)
	{
		return 0;
	}
	return 1;
}

int loadvolumes(volume *volumes, int max)
{
	FILE *fd = setmntent("/etc/mtab", "r");
	if (fd==NULL)
	{
		return -1;
	}
	else
	{
		struct mntent *mnt;
		int vols = 0;
		while ( (mnt=getmntent(fd))!=0)
		{
	 		if (interestingfs(mnt))
			{
				strncpy(volumes[vols].name, mnt->mnt_dir,MAX_VOLUME_NAME);
				volumes[vols].name[MAX_VOLUME_NAME-1]=0;
				volumes[vols].alarmed = -1;
				vols++;
			}
		}
		endmntent(fd);
		return vols;
	}
}

int volumeusage(char *p)
{
	struct statvfs s;
	if (statvfs(p, &s)!=0)
	{
		return -1;
	}
	else
	{
		return percent(s.f_blocks, s.f_bfree, s.f_bavail);
	}
}

static long getvalue(char *p, char *field)
{
	p = strstr(p, field);
	if (p==NULL)
	{
		return -1L;
	}
	while (*p!=' ' && *p!=0) p++;
	while (*p==' ' && *p!=0) p++;
	return *p==0?-1L:strtoul(p,&p,0);
}

static int swapusage(char *buffer)
{
	double total = (double) getvalue(buffer, "SwapTotal:");
	double free = (double) getvalue(buffer, "SwapFree:");

	if (total==-1.0 || free==-1.0) return 0;
        return ((total-free)*100.00)/total;
}
static int pmemusage(char *buffer)
{
	double total = (double) getvalue(buffer, "MemTotal:");
	double free = (double) getvalue(buffer, "MemFree:");
	double buffers = (double) getvalue(buffer, "Buffers:");
	double cached = (double) getvalue(buffer, "Cached:");
	if (total==-1.0 || free==-1.0) return 0; /* the others are not as important */

	free+=(buffers+cached);
        return ((total-free)*100.00)/total;
}

static void memory(int readings[])
{
	char buffer[4096+1];
	int len;

	int fd = open("/proc/meminfo", O_RDONLY);
	if (fd<0)
	{
		iao_logsysmsg("Can't open /proc/meminfo");
		return;
	}
	
	len = read(fd, buffer, sizeof(buffer)-1);
	close(fd);
	if (len<=0)
	{
		iao_logsysmsg("Can't read from /proc/meminfo");
		return;
	}
	buffer[len]=0;

	readings[SWAP] = swapusage(buffer);
	readings[PMEM] = pmemusage(buffer);
}

static long percentages(int cnt, int *out, long long *new, long long *old, long long *diffs)
{
    register int i;
    register long long change;
    register long long total_change;
    register long long *dp;
    long long half_total;

    /* initialization */
    total_change = 0;
    dp = diffs;

    /* calculate changes for each state and the overall change */
    for (i = 0; i < cnt; i++)
    {
        if ((change = *new - *old) < 0)
        {
            /* this only happens when the counter wraps */
            change = (int)
                ((unsigned long long)*new-(unsigned long long)*old);
        }
        total_change += (*dp++ = change);
        *old++ = *new++;
    }

    /* avoid divide by zero potential */
    if (total_change == 0)
    {
        total_change = 1;
    }

    /* calculate percentages based on overall change, rounding up */
    half_total = total_change / 2l;
    for (i = 0; i < cnt; i++)
    {
        *out++ = (int)((*diffs++ * 1000 + half_total) / total_change);
    }

    /* return the total in case the caller wants to use it */
    return(total_change);
}

static int cpusnapshot()
{
	static long long cp_time[NCPUSTATES];
	static long long cp_old[NCPUSTATES];
	static long long cp_diff[NCPUSTATES];
	static int cpu_states[NCPUSTATES];
        char buffer[4096+1];
        int len;
        char *p;

        int fd = open("/proc/stat", O_RDONLY);
        if (fd<0)
        {
		iao_logsysmsg("Can't open /proc/stat");
                return -1;
        }

        len = read(fd, buffer, sizeof(buffer)-1);
        close(fd);
        if (len<=0)
        {
		iao_logsysmsg("Can't read from /proc/stat");
                return -1;
        }
        buffer[len] = 0;
        p = buffer+3;
        cp_time[0] = strtoull(p, &p, 0); /* user */
        cp_time[1] = strtoull(p, &p, 0); /* nice */
        cp_time[2] = strtoull(p, &p, 0); /* system */
        cp_time[3] = strtoull(p, &p, 0); /* idle */

        percentages(NCPUSTATES, cpu_states, cp_time, cp_old, cp_diff);
	return (1000-cpu_states[3])/10;
}

static int cpuusage()
{
        static time_t last = 0;
        time_t now = time(NULL);

	if (last==0)
	{
		if (cpusnapshot()>=0) /* init */
		{
			last = now;
		}
		return 0;
	}
	else
	{
		if (now-last<MIN_CPU_SAMPLE)
		{
			return 0;
		}
		else
		{
			int ret = cpusnapshot();
			if (ret>=0)
			{
				last = now;
        			return ret;
			}
			else
			{
				return 0;
			}
		}
	}
}


int getmetrics(int readings[])
{
	memory(readings);
	readings[CPU] = cpuusage();
	return 0;
}


#endif /* Linux */

#ifdef SunOS

#include <stdio.h>
#include <sys/mnttab.h>
#include <sys/types.h>
#include <sys/statvfs.h>
#include <sys/stat.h>
#include <sys/swap.h>

static int interestingfs(struct mnttab *mnt)
{
	char *dev = "/dev/";
	char *rw = "rw,";

	if (strcmp(mnt->mnt_special, "swap")==0)
	{
		return 1;
	}
        if (strncmp(mnt->mnt_special, dev, strlen(dev))!=0)
        {
                return 0;        
	}
	if (strstr(mnt->mnt_mntopts, rw)==NULL)
	{
		return 0;
	}
	return 1;
}

int loadvolumes(volume *volumes, int max)
{
	FILE *fd = fopen("/etc/mnttab", "r");
	int vols = 0;
	if (fd!=NULL)
	{
		struct mnttab mnt;
		while (getmntent(fd, &mnt)==0)
		{
			if (interestingfs(&mnt))
			{
				strcpy(volumes[vols].name, mnt.mnt_mountp);
				volumes[vols].alarmed = -1;
				vols++;
			}
		}
		fclose(fd);
	}
	return vols;
}

int volumeusage(char *mntp)
{
	struct statvfs s;

	if (statvfs(mntp, &s)==0)
	{
		double total = s.f_blocks-s.f_bfree+s.f_bavail;
		double avail = s.f_bavail;
		int percent = 100 - (int)(avail*100.00/total);
		return percent;
	}
	else
	{
		return -1;
	}
}

static int swapusage()
{
	struct anoninfo anon;
	double total;
	double fr;

	if (swapctl(SC_AINFO, &anon) == -1)
	{
		return -1;
	}
	else
	{
		total = anon.ani_max;
		fr = anon.ani_max - anon.ani_resv;
		return (int)(  ((total-fr)/total)*100.00 );
	}
}

int getmetrics(int readings[])
{
	readings[SWAP] = swapusage();
	return 0;
}


#endif /* SunOS */


/****** HP
#include <stdio.h>
#include <mntent.h>
#include <sys/types.h>
#include <sys/statvfs.h>
#include <math.h>


main()
{
	FILE *fd = fopen("/etc/mnttab", "r");
	if (fd!=NULL)
	{
		struct mntent *mnt;
		struct statvfs s;
		while ( (mnt=getmntent(fd))!=0)
		{
			printf("name=%s mntpoint=%s type=%s",
					 mnt->mnt_fsname, mnt->mnt_dir, mnt->mnt_type);
			if (statvfs(mnt->mnt_dir, &s)==0)
			{
				double total = s.f_blocks-s.f_bfree+s.f_bavail;
				double avail = s.f_bavail;
				int percent = 100 - (int)(avail*100.00/total);
				printf(" %d\n", percent);
			}
			else
			{
				printf(" E\n");
			}
		}
	}
}
*/

